/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <genvector/gds_char.h>
#include <genlist/gendlist.h>
#include <libfungw/fungw.h>
#include "window.h"
#include "edit.h"
#include "conf.h"
#include "fgwirc.h"

#define SCROLL_JUMP 4
#define EDIT_LINE_MAX_LEN 500

gds_t edit_line;
int edit_cursor, edit_scroll;

/*** history ***/
int edit_history_len = 128;

static int history_pos = 0; /* 0=editing the unfinished, -1 is the last entered, -2 is the message before, etc. */

typedef struct {
	gdl_elem_t link;
	gds_t line;
} hist_t;

static hist_t unfinished; /* the line that was being edited when history stepping started */
gdl_list_t finished; /* lines already entered; first is oldest, last is most recent */

static void history_step(int step)
{
	int newpos = history_pos + step, min;
	hist_t *h;

	min = -gdl_length(&finished);
/*	fprintf(stderr, "Ahpos: %d npos: %d bump: %d\n", history_pos, newpos, min);*/

	if (gdl_length(&finished) == 0) return;
	if (newpos > 0) newpos = 0;
	if (newpos < min) newpos = min;
	if (newpos == history_pos) return;

	/* moving away from unfinished edit - save */
	if (history_pos == 0) {
		unfinished.line = edit_line;
		gds_init(&edit_line);
	}

	/* resotre the unfinished */
	if (newpos == 0) {
		gds_uninit(&edit_line);
		edit_line = unfinished.line;
		gds_init(&unfinished.line);
		history_pos = 0;
		edit_scroll = 0;
		edit_cursor = edit_line.used;
		return;
	}

	history_pos = newpos;

	gds_uninit(&edit_line);
	h = finished.last;
	for(newpos++; newpos < 0; newpos++)
		h = h->link.prev;

	edit_line.array = fgw_strdup(h->line.array);
	edit_line.alloced = edit_line.used = h->line.used;
	edit_scroll = 0;
	edit_cursor = edit_line.used;
}

void history_enter()
{
	hist_t *h = calloc(sizeof(hist_t), 1);
	int max;

	if (edit_line.used == 0)
		return;

	h->line.array = fgw_strdup(edit_line.array);
	h->line.alloced = h->line.used = edit_line.used;
	gdl_append(&finished, h, link);

	max = edit_history_len;
	if (max < 1)
		max = 1;

	while(gdl_length(&finished) > max) {
		h = gdl_first(&finished);
		gdl_remove(&finished, h, link);
		gds_uninit(&h->line);
		free(h);
	}

	history_pos = 0;
}

/*** line edit ***/

int edit_draw()
{
	static char spaces[] = "                                                                                                                                                                                                                                                                                                                            ";
	int len, rem;

	tcurses_gotoxy(term, 0, term->sy-1);
	len = edit_line.used - edit_scroll;
	if (len > term->sx-1)
		len = term->sx-1;

	tcurses_write_len(term, edit_line.array + edit_scroll, len);
	rem = term->sx - 1 - len;
	if (rem > 0) {
		if (rem > sizeof(spaces)-1)
			rem = sizeof(spaces)-1;
		tcurses_write_len(term, spaces, rem);
	}
	tcurses_cursor(term, edit_cursor-edit_scroll, term->sy-1);
	return len;
}

int edit_key(input_mod_t modifier, int keycode)
{
	int cp;
	char tmp[256];

	switch(modifier) {
		case IMOD_NONE:
			if ((keycode == 10) || (keycode == 13)) {
				history_enter();
				gds_append(&edit_line, 0);
				fgwirc_enter_str(edit_line.array);
				gds_truncate(&edit_line, 0);
				edit_cursor = 0;
				goto done;
			}
			if ((keycode >= 32) && (keycode < 127)) {
				char c = keycode;
				if (edit_line.used < EDIT_LINE_MAX_LEN) {
					gds_insert_len(&edit_line, edit_cursor, &c, 1);
					edit_cursor++;
				}
				goto done;
			}
			if ((keycode == 8) || (keycode == 127)) { /* backspace */
				if (edit_cursor > 0) {
					gds_remove(&edit_line, edit_cursor-1, 1);
					edit_cursor--;
					goto done;
				}
				return 0;
			}
			break;
		case IMOD_PAD:
			switch(keycode) {
				case 49: /* home */
					edit_cursor = 0;
					goto done;

				case 52: /* end */
					edit_cursor = edit_line.used;
					goto done;

				case 65: /* up */
					history_step(-1);
					goto done;

				case 66: /* down */
					history_step(+1);
					goto done;

				case 68: /* left */
					edit_cursor--;
					if (edit_cursor < 0)
						edit_cursor = 0;
					goto done;

				case 67: /* right */
					edit_cursor++;
					if (edit_cursor >= edit_line.used)
						edit_cursor = edit_line.used;
					goto done;

				case 71: /* numpad 5 -> debug */
					tcurses_dump_to_file(term, "TERM.dump");
					return 1;

				break;
			}

		default:
			break;
	}

	sprintf(tmp, "Unknown key: %d:%d\n", modifier, keycode);
	win_append(wstatus, tmp);

	return 0;

	done:;

	cp = edit_cursor - edit_scroll;
	if (cp < 0) {
		edit_scroll += cp;
		edit_scroll -= SCROLL_JUMP;
		if (edit_scroll < 0)
			edit_scroll = 0;
	}
	if (cp >= term->sx-1) {
		edit_scroll += cp - (term->sx-1);
		edit_scroll += SCROLL_JUMP;
	}

	edit_draw();
	tcurses_flush(term);
	return 1;
}

void edit_init()
{
	conf_reg("edit_history_len", FGW_INT, &edit_history_len, NULL, "number of lines remembered for edit line history (at least 1)");
}
