/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <tcurses/tcurses.h>
#include <libfungw/fungw.h>
#include <iconv.h>

#include "conf.h"
#include "edit.h"
#include "fgwirc.h"

#include "window.h"

tcurses_term_t *term;
int win_max_scroll_back_len = 256;
int win_display_timestamp;
vtp0_t windows;
int win_curr_idx = -1;

void win_dirty(win_t *win)
{
	if (!win->dirty) {
		status_dirty = 1;
		win->dirty = 1;
	}
}

void conf_chg_rewrap(conf_item_t *c)
{
	win_dirty(idx2win(win_curr_idx));
}

static unsigned char hun_latin[] = {
	0xC1, 0xC9, 0xCD, 0xD3, 0xD5, 0xD6, 0xDA, 0xDB, 0xDC, 0xE1,
	0xE9, 0xED, 0xF3, 0xF5, 0xF6, 0xFA, 0xFB, 0xFC, 0};

/* accept iso8859 only if it contains ASCII and hungarian accented characters */
static int is_iso_8859(const char *s, int len)
{
	for(; (*s !='\0') && (len > 0); s++, len--) {
		if (*(unsigned char *)s < 127) continue;
		if (strchr((char *)hun_latin, *s) == NULL)
			return 0;
	}
	return 1;
}

static iconv_t *i_latin = NULL, *i_utf = NULL;

char *win_encoding = "iso-8859-1//TRANSLIT";

void win_init_encoding(void)
{
	if (i_latin != NULL) iconv_close(i_latin);
	if (i_utf != NULL) iconv_close(i_utf);

	i_latin = iconv_open(win_encoding, "iso-8859-2");
	i_utf = iconv_open(win_encoding, "UTF-8");
}

void conf_chg_encoding(conf_item_t *c)
{
	win_init_encoding();
}

void win_resize(int sx, int sy)
{
	win_t *curr = idx2win(win_curr_idx);

	if ((sx == 0) || (sy == 0))
		win_get_screen_size(&sx, &sy);
	tcurses_resize(term, sx, sy);
	curr->dirty = 1;
	status_dirty = 1;
}

static void sigwinch(int signo)
{
	win_resize(0, 0);
}

void win_init(void)
{
	int sx, sy;
	win_get_screen_size(&sx, &sy);
	term = tcurses_init(1, "vt100", sx, sy);
	if (term == NULL) {
		fprintf(stderr, "failed to initialize the terminal.\n");
		exit(1);
	}
	httw_init(&target2win, httw_keyhash, httw_keyeq);

	conf_reg("win_max_scroll_back_len", FGW_INT, &win_max_scroll_back_len, NULL, "Change the size of the scroll-back buffer (number of raw, unwrapped lines); global to all windows; takes effect in each window the next time a line is appended");
	conf_reg("win_display_timestamp", FGW_INT, &win_display_timestamp, conf_chg_rewrap, "Whether to display time stamp with each raw line on screen (does not affect logging)");
	conf_reg("win_encoding", FGW_STR, &win_encoding, conf_chg_encoding, "Convert incoming messages to the specified character encoding before inserting in the window log (expects iconv(3) encoding name)");

	signal(SIGWINCH, sigwinch);
	signal(SIGINT,  SIG_IGN);

}

void win_uninit()
{
	tcurses_uninit(term);
	httw_uninit(&target2win);
}

static int win_sx(void)
{
	if (win_display_timestamp)
		return term->sx - 9;

	return term->sx;
}

win_t *idx2win(int idx)
{
	void **win;
	win = vtp0_get(&windows, idx, 0);
	if ((win == NULL) || (*win == NULL))
		return NULL;
	return *win;
}


wline_t *win_alloc_line(const char *line, int len, int do_encoding)
{
	size_t li, lo = len*4;
	wline_t *l;
	char *si;
	char *so;

	if (do_encoding) {
		iconv_t *i;

		if (i_latin == NULL)
			win_init_encoding();

		l = malloc(sizeof(wline_t) + lo);
		so = l->s;
		si = (char *)line;
		li = len;

		if (is_iso_8859(line, len))
			i = i_latin;
		else
			i = i_utf;

		for(;;) {
			iconv(i, &si, &li, &so, &lo);
			if ((*si == 0) || (li == 0))
				break;
#if 0
			can not use this because this makes the output longer than the input
			and interferes with wrapping
			if (lo > 4) {
				int len;
				*so = '\\'; so++; lo--;
				len = sprintf(so, "%02x", *(unsigned char *)si);
				so += len; lo -= len;
			}
			else
#endif
			if (lo > 2) {
				*so = '?';
				lo--;
			}
			si++;
		}
		*so = '\0';
		l->len = so - l->s;
	}
	else {
		l = malloc(sizeof(wline_t) + len);
		memcpy(l->s, line, len);
		l->s[len] = '\0';
	l->len = len;
	}
	memset(&l->link, 0, sizeof(gdl_elem_t));
	return l;
}

void win_wrap_append(win_t *win, wline_t *raw)
{
	char *start, *s, *spc = NULL;
	wline_t *l;
	int first = 1;

	for(start = s = raw->s; *s != '\0'; s++) {
		int len = s - start;

		if (len >= win->wrap_sx) {
			wline_t *l;
			char *to;

			if (spc != NULL)
				to = spc;
			else
				to = s;
			l = win_alloc_line(start, to-start, 1);
			l->raw = raw;
			l->first = first;
			first = 0;
			l->time = raw->time;
			gdl_append(&win->wrapped, l, link);
			start = s = to;
			spc = NULL;
		}
		else if ((*s == ' ') || (*s == '\t'))
			spc = s;
	}

	l = win_alloc_line(start, s-start, 1);
	l->raw = raw;
	l->first = first;
	l->time = raw->time;
	gdl_append(&win->wrapped, l, link);
}

void win_rewrap(win_t *win, int sx)
{
	wline_t *l;

	win->wrap_sx = sx;

	/* remove all wrapped */
	for(l = win->wrapped.first; l != NULL; l = win->wrapped.first) {
		gdl_remove(&win->wrapped, l, link);
		free(l);
	}

	/* do the new wrapping */
	for(l = win->raw.first; l != NULL; l = l->link.next)
		win_wrap_append(win, l);
}

void win_render_time(char *dst, const struct tm *tm, int bracket)
{
	if (bracket) {
		if (tm == NULL)
			strcpy(dst, "         ");
		else
			strftime(dst, 9, "[%H:%M] ", tm);
	}
	else {
		if (tm == NULL)
			strcpy(dst, "      ");
		else
			strftime(dst, 6, "%H:%M", tm);
	}
}


void win_append_(win_t *win, const char *line)
{
	const char *curr, *next;
	wline_t *l = NULL;
	int len;

	if (win->wrap_sx != win_sx())
		win_rewrap(win, win_sx());

	for(curr = line; curr != NULL; curr = next) {
		next = strchr(curr, '\n');
		if (next != NULL) {
			len = next - curr;
			next++;
			if (*next == '\0')
				next = NULL;
		}
		else
			len = strlen(line);

		l = win_alloc_line(curr, len, 0);
		l->time = now;
		gdl_append(&win->raw, l, link);
		win_wrap_append(win, l);

		if (win->logf != NULL) {
			char tmp[16];
			win_render_time(tmp, localtime(&l->time), 1);
			fprintf(win->logf, "%s %s\n", tmp, l->s);
			win->needs_log_flush = 1;
		}
	}
}

void win_append(win_t *win, const char *line)
{
	win_append_(win, line);
	if (win->scroll_start == NULL) {
		while(win->raw.length > win_max_scroll_back_len)
			win_del_oldest(win);
	}
	win_dirty(win);
}

void win_draw_status(win_t *win)
{
	char tim[16], tmp[256], *s, *end;
	int n, max, len, sx = term->sx, first;

	win_render_time(tim, &now_tm, 0);
	len = sprintf(tmp, "[% 2d] %s ", win_curr_idx, tim);

	/* activity */
	first = 1;
	for(n = 0; n <  windows.used; n++) {
		win_t *w = windows.array[n];
		if (n == win->idx) continue;
		if ((w != NULL) && (w->dirty)) {
			end = tmp + len;
			len += sprintf(end, "%s%d", first ? "|" : ",", n);
			first = 0;
		}
	}

	/* scrollback */
	if (win->scroll_start != NULL) {
		end = tmp + len;
		len += sprintf(end, "|S%d", win->scroll_pending);
	}

	/* current window info */
	end = tmp + len;
	len += sprintf(end, "| %s", win->name);
	if (win->target.sid >= 0) {
		end = tmp + len;
		len += sprintf(end, " on %d", win->target.sid);
	}

	if (len < sx) {
		max = sx - len;
		if (max > sizeof(tmp)-2)
			max = sizeof(tmp)-2;
		s = tmp+len;
		for(n = 0; n < max; n++,s++)
			*s = ' ';
		*s = '\0';
	}
	else
		tmp[sx] = '\0';

	term->attr = 7;
	tcurses_gotoxy(term, 0, term->sy-2);
	tcurses_write(term, tmp);
	term->attr = 0;
}

void win_flush(void)
{
	tcurses_flush(term);
}

void win_draw(win_t *win)
{
	int y, sx;
	wline_t *l;

	sx = win_sx();

	tcurses_clear(term);
	if (win->wrap_sx != sx)
		win_rewrap(win, sx);

	if (win->scroll_start != NULL)
		l = win->scroll_start;
	else
		l = win->wrapped.last;


	/* draw lines */
	for(y = term->sy-3; (l != NULL) && (y >= 0); y--) {
		if (win_display_timestamp) {
			if (l->first) {
				char tim[16];
				win_render_time(tim, localtime(&l->time), 1);
				tcurses_gotoxy(term, 0, y);
				tcurses_write(term, tim);
			}
			else
				tcurses_gotoxy(term, 8, y);
		}
		else
			tcurses_gotoxy(term, 0, y);
		tcurses_write(term, l->s);
		l = l->link.prev;
	}

	win_draw_status(win);
	edit_draw();
	win_flush();
	win->dirty = 0;
}

void win_scroll_up(win_t *win, int lines)
{
	if (win->scroll_start == NULL)
		win->scroll_start = win->wrapped.last;

	for(; (lines > 0) && (win->scroll_start->link.prev != NULL); lines--) {
		win->scroll_start = win->scroll_start->link.prev;
		win->scroll_pending++;
	}
}

void win_scroll_down(win_t *win, int lines)
{
	if (win->scroll_start == NULL)
		return;

	for(; (lines > 0) && (win->scroll_start != NULL); lines--) {
		win->scroll_start = win->scroll_start->link.next;
		win->scroll_pending--;
	}
}

void win_del_oldest(win_t *win)
{
	wline_t *raw = win->raw.first, *l;

	if (raw == NULL)
		return;

	/* remove wrapped version */
	for(l = win->wrapped.first; (l != NULL) && (l->raw == raw); l = win->wrapped.first) {
		gdl_remove(&win->wrapped, l, link);
		free(l);
	}

	gdl_remove(&win->raw, raw, link);
	free(raw);
}

const char *win_log(win_t *win, const char *name, int on)
{
	char tmp[256];

	if (on) {
		if (win->logf != NULL)
			return "logging already on";
		if (name == NULL) {
			sprintf(tmp, "Irclog.window#%d", win->idx);
			name = tmp;
		}
		win->logf = fopen(name, "a");
		if (win->logf == NULL)
			return "failed to open log file";
	}
	else {
		if (win->logf == NULL)
			return "logging was not on";
		fclose(win->logf);
		win->logf = NULL;
		win->needs_log_flush = 0;
	}

	return NULL;
}

win_t *win_new(const char *name)
{
	win_t *win = calloc(sizeof(win_t), 1);

	vtp0_append(&windows, win);
	win->idx = windows.used-1;
	win->name = fgw_strdup(name);

	{
		target_t *target = (target_t *)&win->target; /* this is one of the three legitime places to write these data directly */
		target->type = TARGET_NONE;
		target->sid = -1;
	}

	return win;
}

void win_show_idx(int idx)
{
	void **win;
	if (win_curr_idx == idx)
		return;
	win = vtp0_get(&windows, idx, 0);
	if ((win == NULL) || (*win == NULL))
		return;
	win_curr_idx = idx;
	win_draw(*win);
}

void win_show(win_t *win)
{
	win_show_idx(win->idx);
}

const char win_target_keep[] = "keep";

void win_set_target(win_t *win, target_type_t type, int sid, const char *name, const char *key)
{
	httw_entry_t *user;
	target_t *target = (target_t *)&win->target; /* this is one of the three legitime places to write these data directly */
	char *end;
	long nsid;

	httw_pop(&target2win, (target_t *)&win->target);

	/* override sid if name is N/ prefixed */
	nsid = strtol(name, &end, 10);
	if (*end == '/') {
		sid = nsid;
		name = end+1;
	}

	target->type = type;
	target->sid = sid;
	if (name != win_target_keep) {
		free(target->name);
		target->name = fgw_strdup(name);
	}
	if (key != win_target_keep) {
		free(target->key);
		if (key != NULL)
			target->key = fgw_strdup(key);
		else
			target->key = NULL;
	}

	/* remove other window using the same target */
	if (target->type != TARGET_NONE) {
		user = httw_popentry(&target2win, (target_t *)&win->target);
		if (user != NULL) {
			target->type = TARGET_NONE;
			/* keep sid */
			free(target->name);
			target->name = NULL;
			free(win->name);
			win->name = fgw_strdup("-");
		}
		httw_set(&target2win, (target_t *)&win->target, win);
	}
}

void win_set_target_state(win_t *win, target_state_t st)
{
	target_t *target = (target_t *)&win->target; /* this is one of the three legitime places to write these data directly */
	target->state = st;
}

void win_set_target_sid(win_t *win, int sid)
{
	win_set_target(win, win->target.type, sid, win_target_keep, win_target_keep);
}


win_t *win_target2win(target_type_t type, int sid, const char *name)
{
	target_t t;
	t.type = type;
	t.sid = sid;
	t.name = (char *)name; /* safe: hashing won't change the key */
	return httw_get(&target2win, &t);
}


void win_get_screen_size(int *sx, int *sy)
{
	struct winsize ws;
	if (ioctl(0, TIOCGWINSZ, &ws) == 0) {
		*sx = ws.ws_col;
		*sy = ws.ws_row;
	}
	else {
		*sx = 80;
		*sy = 25;
	}
}

